@@ -50,11 +50,20 @@
 #include "error.h"
 #include "socket.h"
 #include "setspeed.h"
+#include "rs485.h"
+#include "alert.h"
+#include "timestamp.h"
+#include "misc.h"
 
-#ifdef __APPLE__
+#if defined(__APPLE__)
 #define PATH_SERIAL_DEVICES "/dev/"
+#define PREFIX_TTY_DEVICES "tty."
+#elif defined(__CYGWIN__)
+#define PATH_SERIAL_DEVICES "/dev/"
+#define PREFIX_TTY_DEVICES "ttyS"
 #else
 #define PATH_SERIAL_DEVICES "/dev/serial/by-id/"
+#define PREFIX_TTY_DEVICES ""
 #endif
 
 #ifndef CMSPAR
@@ -74,16 +83,15 @@
 #define KEY_G 0x67
 #define KEY_H 0x68
 #define KEY_L 0x6C
+#define KEY_SHIFT_L 0x4C
+#define KEY_M 0x6D
 #define KEY_P 0x70
 #define KEY_Q 0x71
 #define KEY_S 0x73
 #define KEY_T 0x74
 #define KEY_U 0x55
 #define KEY_V 0x76
-#define KEY_SHIFT_L 0x4C
-
-#define NORMAL 0
-#define HEX 1
+#define KEY_Z 0x7a
 
 enum line_mode_t
 {
@@ -92,12 +100,25 @@ enum line_mode_t
     LINE_PULSE
 };
 
+const char random_array[] =
+{
+0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x28, 0x20, 0x28, 0x0A, 0x20,
+0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x29, 0x20, 0x29, 0x0A, 0x20,
+0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x2E, 0x2E, 0x2E, 0x2E, 0x2E, 0x2E, 0x2E,
+0x2E, 0x0A, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x7C, 0x20, 0x20, 0x20,
+0x20, 0x20, 0x20, 0x7C, 0x5D, 0x0A, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
+0x5C, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x2F, 0x0A, 0x20, 0x20, 0x20, 0x20,
+0x20, 0x20, 0x20, 0x20, 0x60, 0x2D, 0x2D, 0x2D, 0x2D, 0x27, 0x0A, 0x0A, 0x54,
+0x69, 0x6D, 0x65, 0x20, 0x66, 0x6F, 0x72, 0x20, 0x61, 0x20, 0x63, 0x6F, 0x66,
+0x66, 0x65, 0x65, 0x20, 0x62, 0x72, 0x65, 0x61, 0x6B, 0x21, 0x0A, 0x20, 0x0A,
+0x00
+};
+
 bool interactive_mode = true;
 
 static struct termios tio, tio_old, stdout_new, stdout_old, stdin_new, stdin_old;
 static unsigned long rx_total = 0, tx_total = 0;
 static bool connected = false;
-static bool print_mode = NORMAL;
 static bool standard_baudrate = true;
 static void (*print)(char c);
 static int fd;
@@ -106,6 +127,7 @@ static bool map_o_cr_nl = false;
 static bool map_o_nl_crnl = false;
 static bool map_o_del_bs = false;
 static bool map_o_ltu = false;
+static bool map_o_msblsb = false;
 static char hex_chars[2];
 static unsigned char hex_char_index = 0;
 static char tty_buffer[BUFSIZ*2];
@@ -201,7 +223,7 @@ ssize_t tty_write(int fd, const void *buffer, size_t count)
             }
             bytes_written += retval;
 
-            if (option.output_line_delay && *(unsigned char*)buffer == '\r')
+            if (option.output_line_delay && *(unsigned char*)buffer == '\n')
             {
                 delay(option.output_line_delay);
             }
@@ -237,8 +259,14 @@ static void output_hex(char c)
 {
     hex_chars[hex_char_index++] = c;
 
+    printf("%c", c);
+
     if (hex_char_index == 2)
     {
+        usleep(100*1000);
+        printf("\b \b");
+        printf("\b \b");
+
         unsigned char hex_value = char_to_nibble(hex_chars[0]) << 4 | (char_to_nibble(hex_chars[1]) & 0x0F);
         hex_char_index = 0;
 
@@ -382,20 +410,22 @@ void handle_command_sequence(char input_char, char previous_char, char *output_c
         {
             case KEY_QUESTION:
                 tio_printf("Key commands:");
-                tio_printf(" ctrl-%c ?   List available key commands", option.prefix_key);
-                tio_printf(" ctrl-%c b   Send break", option.prefix_key);
-                tio_printf(" ctrl-%c c   Show configuration", option.prefix_key);
-                tio_printf(" ctrl-%c e   Toggle local echo mode", option.prefix_key);
-                tio_printf(" ctrl-%c g   Toggle serial port line", option.prefix_key);
-                tio_printf(" ctrl-%c h   Toggle hexadecimal mode", option.prefix_key);
-                tio_printf(" ctrl-%c l   Clear screen", option.prefix_key);
-                tio_printf(" ctrl-%c L   Show line states", option.prefix_key);
-                tio_printf(" ctrl-%c p   Pulse serial port line", option.prefix_key);
-                tio_printf(" ctrl-%c q   Quit", option.prefix_key);
-                tio_printf(" ctrl-%c s   Show statistics", option.prefix_key);
-                tio_printf(" ctrl-%c t   Toggle line timestamp mode", option.prefix_key);
-                tio_printf(" ctrl-%c U   Toggle conversion to uppercase on output", option.prefix_key);
-                tio_printf(" ctrl-%c v   Show version", option.prefix_key);
+                tio_printf(" ctrl-%c ?       List available key commands", option.prefix_key);
+                tio_printf(" ctrl-%c b       Send break", option.prefix_key);
+                tio_printf(" ctrl-%c c       Show configuration", option.prefix_key);
+                tio_printf(" ctrl-%c e       Toggle local echo mode", option.prefix_key);
+                tio_printf(" ctrl-%c g       Toggle serial port line", option.prefix_key);
+                tio_printf(" ctrl-%c h       Toggle hexadecimal mode", option.prefix_key);
+                tio_printf(" ctrl-%c l       Clear screen", option.prefix_key);
+                tio_printf(" ctrl-%c L       Show line states", option.prefix_key);
+                tio_printf(" ctrl-%c m       Toggle MSB to LSB bit order", option.prefix_key);
+                tio_printf(" ctrl-%c p       Pulse serial port line", option.prefix_key);
+                tio_printf(" ctrl-%c q       Quit", option.prefix_key);
+                tio_printf(" ctrl-%c s       Show statistics", option.prefix_key);
+                tio_printf(" ctrl-%c t       Toggle line timestamp mode", option.prefix_key);
+                tio_printf(" ctrl-%c U       Toggle conversion to uppercase on output", option.prefix_key);
+                tio_printf(" ctrl-%c v       Show version", option.prefix_key);
+                tio_printf(" ctrl-%c ctrl-%c  Send ctrl-%c character", option.prefix_key, option.prefix_key, option.prefix_key);
                 break;
 
             case KEY_SHIFT_L:
@@ -445,6 +475,10 @@ void handle_command_sequence(char input_char, char previous_char, char *output_c
                 tio_printf("Configuration:");
                 config_file_print();
                 options_print();
+                if (option.rs485)
+                {
+                    rs485_print_config();
+                }
                 break;
 
             case KEY_E:
@@ -454,16 +488,16 @@ void handle_command_sequence(char input_char, char previous_char, char *output_c
 
             case KEY_H:
                 /* Toggle hexadecimal printing mode */
-                if (print_mode == NORMAL)
+                if (!option.hex_mode)
                 {
                     print = print_hex;
-                    print_mode = HEX;
+                    option.hex_mode = true;
                     tio_printf("Switched to hexadecimal mode");
                 }
                 else
                 {
                     print = print_normal;
-                    print_mode = NORMAL;
+                    option.hex_mode = false;
                     tio_printf("Switched to normal mode");
                 }
                 break;
@@ -473,6 +507,20 @@ void handle_command_sequence(char input_char, char previous_char, char *output_c
                 printf("\033c");
                 break;
 
+            case KEY_M:
+                /* Toggle bit order */
+                if (!map_o_msblsb)
+                {
+                    map_o_msblsb = true;
+                    tio_printf("Switched to reverse bit order");
+                }
+                else
+                {
+                    map_o_msblsb = false;
+                    tio_printf("Switched to normal bit order");
+                }
+                break;
+
             case KEY_Q:
                 /* Exit upon ctrl-t q sequence */
                 exit(EXIT_SUCCESS);
@@ -517,7 +565,28 @@ void handle_command_sequence(char input_char, char previous_char, char *output_c
                 tio_printf("tio v%s", VERSION);
                 break;
 
+            case KEY_Z:
+                tio_printf_array(random_array);
+                break;
+
             default:
+                /* Handle double prefix key input case */
+                if (input_char == option.prefix_code)
+                {
+                    static int count = 0;
+                    if (count++ == 1)
+                    {
+                        // Do not forward prefix characters excessively
+                        count = 0;
+                        break;
+                    }
+
+                    /* Forward prefix character to tty */
+                    *output_char = option.prefix_code;
+                    *forward = true;
+                    break;
+                }
+
                 /* Ignore unknown ctrl-t escaped keys */
                 break;
         }
@@ -814,6 +883,10 @@ void tty_configure(void)
             {
                 map_o_ltu = true;
             }
+            else if (strcmp(token, "MSB2LSB") == 0)
+            {
+                map_o_msblsb = true;
+            }
             else
             {
                 printf("Error: Unknown mapping flag %s\n", token);
@@ -925,6 +998,9 @@ void tty_disconnect(void)
         flock(fd, LOCK_UN);
         close(fd);
         connected = false;
+
+        /* Fire alert action */
+        alert_disconnect();
     }
 }
 
@@ -932,6 +1008,12 @@ void tty_restore(void)
 {
     tcsetattr(fd, TCSANOW, &tio_old);
 
+    if (option.rs485)
+    {
+        /* Restore original RS-485 mode */
+        rs485_mode_restore(fd);
+    }
+
     if (connected)
     {
         tty_disconnect();
@@ -969,7 +1051,7 @@ void forward_to_tty(int fd, char output_char)
     }
     else
     {
-        if (print_mode == HEX)
+        if (option.hex_mode)
         {
             output_hex(output_char);
         }
@@ -1000,6 +1082,9 @@ int tty_connect(void)
     int    status;
     bool   next_timestamp = false;
     char*  now = NULL;
+    struct timeval tv;
+    struct timeval *tv_p = &tv;
+    bool ignore_stdin = false;
 
     /* Open tty device */
     fd = open(option.tty_device, O_RDWR | O_NOCTTY | O_NONBLOCK);
@@ -1032,6 +1117,9 @@ int tty_connect(void)
     connected = true;
     print_tainted = false;
 
+    /* Fire alert action */
+    alert_connect();
+
     if (option.timestamp)
     {
         next_timestamp = true;
@@ -1041,12 +1129,10 @@ int tty_connect(void)
     if (option.hex_mode)
     {
         print = print_hex;
-        print_mode = HEX;
     }
     else
     {
         print = print_normal;
-        print_mode = NORMAL;
     }
 
     /* Save current port settings */
@@ -1065,6 +1151,12 @@ int tty_connect(void)
     }
 #endif
 
+    /* Manage RS-485 mode */
+    if (option.rs485)
+    {
+        rs485_mode_enable(fd);
+    }
+
     /* Make sure we restore tty settings on exit */
     if (first)
     {
@@ -1095,12 +1187,28 @@ int tty_connect(void)
     {
         FD_ZERO(&rdfs);
         FD_SET(fd, &rdfs);
-        FD_SET(STDIN_FILENO, &rdfs);
+        if (!ignore_stdin)
+        {
+            FD_SET(STDIN_FILENO, &rdfs);
+        }
         maxfd = MAX(fd, STDIN_FILENO);
         maxfd = MAX(maxfd, socket_add_fds(&rdfs, true));
 
+        /* Manage timeout */
+        if ((option.response_wait) && (option.response_timeout != 0))
+        {
+            // Set response timeout
+            tv_p->tv_sec = 0;
+            tv_p->tv_usec = option.response_timeout * 1000;
+        }
+        else
+        {
+            // No timeout
+            tv_p = NULL;
+        }
+
         /* Block until input becomes available */
-        status = select(maxfd + 1, &rdfs, NULL, NULL, NULL);
+        status = select(maxfd + 1, &rdfs, NULL, NULL, tv_p);
         if (status > 0)
         {
             bool forward = false;
@@ -1124,9 +1232,9 @@ int tty_connect(void)
                     input_char = input_buffer[i];
 
                     /* Print timestamp on new line if enabled */
-                    if (next_timestamp && input_char != '\n' && input_char != '\r')
+                    if ((next_timestamp && input_char != '\n' && input_char != '\r') && !option.hex_mode)
                     {
-                        now = current_time();
+                        now = timestamp_current_time();
                         if (now)
                         {
                             ansi_printf_raw("[%s] ", now);
@@ -1138,8 +1246,19 @@ int tty_connect(void)
                         }
                     }
 
+                    /* Convert MSB to LSB bit order */
+                    if (map_o_msblsb)
+                    {
+                        char ch = input_char;
+                        input_char = 0;
+                        for (int j = 0; j < 8; ++j)
+                        {
+                            input_char |= ((1 << j) & ch) ? (1 << (7 - j)) : 0;
+                        }
+                    }
+
                     /* Map input character */
-                    if ((input_char == '\n') && (map_i_nl_crnl))
+                    if ((input_char == '\n') && (map_i_nl_crnl) && (!map_o_msblsb))
                     {
                         print('\r');
                         print('\n');
@@ -1168,6 +1287,15 @@ int tty_connect(void)
                     {
                         next_timestamp = true;
                     }
+
+                    if (option.response_wait)
+                    {
+                        if ((input_char == '\r') || (input_char == '\n'))
+                        {
+                             tty_sync(fd);
+                             exit(EXIT_SUCCESS);
+                        }
+                    }
                 }
             }
             else if (FD_ISSET(STDIN_FILENO, &rdfs))
@@ -1182,8 +1310,21 @@ int tty_connect(void)
                 else if (bytes_read == 0)
                 {
                     /* Reached EOF (when piping to stdin) */
-                    tty_sync(fd);
-                    exit(EXIT_SUCCESS);
+                    if (option.response_wait)
+                    {
+                        /* Stdin pipe closed but not blocking so stop listening
+                         * to stdin in response mode.
+                         *
+                         * Note: select() really indicates not if data is ready
+                         * but if file descriptor is non-blocking for I/O
+                         * operation. */
+                        ignore_stdin = true;
+                    }
+                    else
+                    {
+                        tty_sync(fd);
+                        exit(EXIT_SUCCESS);
+                    }
                 }
 
                 /* Process input byte by byte */
@@ -1209,7 +1350,7 @@ int tty_connect(void)
                         /* Save previous key */
                         previous_char = input_char;
 
-                        if ((print_mode == HEX) && (forward))
+                        if ((option.hex_mode) && (forward))
                         {
                             if (!is_valid_hex(input_char))
                             {
@@ -1244,6 +1385,11 @@ int tty_connect(void)
             tio_error_printf("select() failed (%s)", strerror(errno));
             exit(EXIT_FAILURE);
         }
+        else
+        {
+            // Timeout (only happens in response wait mode)
+            exit(EXIT_FAILURE);
+        }
     }
 
     return TIO_SUCCESS;
@@ -1267,11 +1413,10 @@ void list_serial_devices(void)
         {
             if ((strcmp(dir->d_name, ".")) && (strcmp(dir->d_name, "..")))
             {
-#ifdef __APPLE__
-#define TTY_DEVICES_PREFIX "tty."
-                if (!strncmp(dir->d_name, TTY_DEVICES_PREFIX, sizeof(TTY_DEVICES_PREFIX) - 1))
-#endif
-                printf("%s%s\n", PATH_SERIAL_DEVICES, dir->d_name);
+                if (!strncmp(dir->d_name, PREFIX_TTY_DEVICES, sizeof(PREFIX_TTY_DEVICES) - 1))
+                {
+                    printf("%s%s\n", PATH_SERIAL_DEVICES, dir->d_name);
+                }
             }
         }
         closedir(d);
